@@ -2,7 +2,8 @@ DROP FUNCTION IF EXISTS last_insert_grn_id;
 DROP FUNCTION IF EXISTS mroonga_snippet;
 DROP FUNCTION IF EXISTS mroonga_command;
 DROP FUNCTION IF EXISTS mroonga_escape;
+DROP FUNCTION IF EXISTS mroonga_snippet_html;
 
-UNINSTALL PLUGIN mroonga;
+UNINSTALL PLUGIN Mroonga;
 
 FLUSH TABLES;
