@@ -32,6 +32,8 @@ extern "C" {
 
 #include <groonga.h>
 #include "mrn_mysql_compat.h"
+#include <mrn_operations.hpp>
+#include <mrn_database.hpp>
 
 #if (MYSQL_VERSION_ID >= 50514 && MYSQL_VERSION_ID < 50600)
 #  define MRN_HANDLER_HAVE_FINAL_ADD_INDEX 1
@@ -204,6 +206,18 @@ extern "C" {
 #  define MRN_HANDLER_HAVE_RESET_AUTO_INCREMENT
 #endif
 
+#if MYSQL_VERSION_ID >= 50709 && !defined(MRN_MARIADB_P)
+#  define MRN_ALTER_INPLACE_INFO_ALTER_STORED_COLUMN_TYPE \
+  Alter_inplace_info::ALTER_STORED_COLUMN_TYPE
+#  define MRN_ALTER_INPLACE_INFO_ALTER_STORED_COLUMN_ORDER \
+  Alter_inplace_info::ALTER_STORED_COLUMN_ORDER
+#else
+#  define MRN_ALTER_INPLACE_INFO_ALTER_STORED_COLUMN_TYPE \
+  Alter_inplace_info::ALTER_COLUMN_TYPE
+#  define MRN_ALTER_INPLACE_INFO_ALTER_STORED_COLUMN_ORDER \
+  Alter_inplace_info::ALTER_COLUMN_ORDER
+#endif
+
 class ha_mroonga;
 
 /* structs */
@@ -347,6 +361,8 @@ private:
   // for ft in where clause test
   Item_func_match *current_ft_item;
 
+  mrn::Operations *operations_;
+
 public:
   ha_mroonga(handlerton *hton, TABLE_SHARE *share_arg);
   ~ha_mroonga();
@@ -617,7 +633,6 @@ private:
   bool is_dry_write();
   bool is_enable_optimization();
   bool should_normalize(Field *field) const;
-  bool is_temporary_table_name(const char *name) const;
   void check_count_skip(key_part_map start_key_part_map,
                         key_part_map end_key_part_map, bool fulltext);
   bool is_grn_zero_column_value(grn_obj *column, grn_obj *value);
@@ -647,6 +662,9 @@ private:
   int generic_store_bulk_new_decimal(Field *field, grn_obj *buf);
   int generic_store_bulk_blob(Field *field, grn_obj *buf);
   int generic_store_bulk_geometry(Field *field, grn_obj *buf);
+#ifdef MRN_HAVE_MYSQL_TYPE_JSON
+  int generic_store_bulk_json(Field *field, grn_obj *buf);
+#endif
   int generic_store_bulk(Field *field, grn_obj *buf);
 
   void storage_store_field_string(Field *field,
@@ -682,6 +700,10 @@ private:
                                 const char *value, uint value_length);
   void storage_store_field_geometry(Field *field,
                                     const char *value, uint value_length);
+#ifdef MRN_HAVE_MYSQL_TYPE_JSON
+  void storage_store_field_json(Field *field,
+                                const char *value, uint value_length);
+#endif
   void storage_store_field(Field *field, const char *value, uint value_length);
   void storage_store_field_column(Field *field, bool is_primary_key,
                                   int nth_column, grn_id record_id);
@@ -762,9 +784,7 @@ private:
                                grn_obj **index_tables,
                                grn_obj **index_columns,
                                MRN_SHARE *tmp_share);
-  int wrapper_create_index(const char *name, TABLE *table,
-                           HA_CREATE_INFO *info, MRN_SHARE *tmp_share,
-                           const char *grn_table_name);
+  int wrapper_create_index(const char *name, TABLE *table, MRN_SHARE *tmp_share);
   int storage_create_validate_pseudo_column(TABLE *table);
 #ifdef MRN_SUPPORT_FOREIGN_KEYS
   bool storage_create_foreign_key(TABLE *table, const char *grn_table_name,
@@ -782,13 +802,14 @@ private:
   int storage_create_indexes(TABLE *table, const char *grn_table_name,
                              grn_obj *grn_table, MRN_SHARE *tmp_share);
   int close_databases();
-  int ensure_database_open(const char *name);
+  int ensure_database_open(const char *name, mrn::Database **db=NULL);
   int ensure_database_remove(const char *name);
   int wrapper_delete_table(const char *name, handlerton *wrap_handlerton,
                            const char *table_name);
   int generic_delete_table(const char *name, const char *table_name);
   int wrapper_open(const char *name, int mode, uint open_options);
   int wrapper_open_indexes(const char *name);
+  int storage_reindex();
   int storage_open(const char *name, int mode, uint open_options);
   int open_table(const char *name);
   int storage_open_columns(void);
@@ -1066,8 +1087,10 @@ private:
   bool wrapper_auto_repair(int error) const;
   bool storage_auto_repair(int error) const;
   int generic_disable_index(int i, KEY *key_info);
+  int wrapper_disable_indexes_mroonga(uint mode);
   int wrapper_disable_indexes(uint mode);
   int storage_disable_indexes(uint mode);
+  int wrapper_enable_indexes_mroonga(uint mode);
   int wrapper_enable_indexes(uint mode);
   int storage_enable_indexes(uint mode);
   int wrapper_check(THD* thd, HA_CHECK_OPT* check_opt);
