@@ -2,7 +2,7 @@
 /*
   Copyright(C) 2010 Tetsuro IKEDA
   Copyright(C) 2010-2013 Kentoku SHIBA
-  Copyright(C) 2011-2013 Kouhei Sutou <kou@clear-code.com>
+  Copyright(C) 2011-2015 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -24,12 +24,20 @@
 #include <mrn_path_mapper.hpp>
 #include <mrn_windows.hpp>
 #include <mrn_macro.hpp>
+#include <mrn_database_manager.hpp>
+#include <mrn_context_pool.hpp>
+#include <mrn_variables.hpp>
 
 MRN_BEGIN_DECLS
 
+extern mrn::DatabaseManager *mrn_db_manager;
+extern mrn::ContextPool *mrn_context_pool;
+
 struct CommandInfo
 {
-  grn_ctx ctx;
+  grn_ctx *ctx;
+  grn_obj *db;
+  bool use_shared_db;
   String result;
 };
 
@@ -53,30 +61,36 @@ MRN_API my_bool mroonga_command_init(UDF_INIT *initid, UDF_ARGS *args,
   initid->maybe_null = 1;
   initid->const_item = 1;
 
-  info = (CommandInfo *)my_malloc(sizeof(CommandInfo),
-                                  MYF(MY_WME | MY_ZEROFILL));
+  info = (CommandInfo *)mrn_my_malloc(sizeof(CommandInfo),
+                                      MYF(MY_WME | MY_ZEROFILL));
   if (!info) {
     strcpy(message, "mroonga_command(): out of memory");
     goto error;
   }
 
-  grn_ctx_init(&(info->ctx), 0);
+  info->ctx = mrn_context_pool->pull();
   {
-    const char *current_db_path = current_thd->db;
+    const char *current_db_path = MRN_THD_DB_PATH(current_thd);
     const char *action;
     if (current_db_path) {
       action = "open database";
-      mrn::PathMapper mapper(current_db_path);
-      grn_db_open(&(info->ctx), mapper.db_path());
+      mrn::Database *db;
+      int error = mrn_db_manager->open(current_db_path, &db);
+      if (error == 0) {
+        info->db= db->get();
+        grn_ctx_use(info->ctx, info->db);
+        info->use_shared_db = true;
+      }
     } else {
       action = "create anonymous database";
-      grn_db_create(&(info->ctx), NULL, NULL);
+      info->db = grn_db_create(info->ctx, NULL, NULL);
+      info->use_shared_db = false;
     }
-    if (info->ctx.rc != GRN_SUCCESS) {
+    if (!info->db) {
       sprintf(message,
               "mroonga_command(): failed to %s: %s",
               action,
-              info->ctx.errbuf);
+              info->ctx->errbuf);
       goto error;
     }
   }
@@ -87,13 +101,11 @@ MRN_API my_bool mroonga_command_init(UDF_INIT *initid, UDF_ARGS *args,
 
 error:
   if (info) {
-    grn_obj *db;
-    db = grn_ctx_db(&(info->ctx));
-    if (db) {
-      grn_obj_close(&(info->ctx), db);
+    if (!info->use_shared_db) {
+      grn_obj_close(info->ctx, info->db);
     }
-    grn_ctx_fin(&(info->ctx));
-    my_free(info, MYF(0));
+    mrn_context_pool->release(info->ctx);
+    my_free(info);
   }
   return TRUE;
 }
@@ -102,7 +114,7 @@ MRN_API char *mroonga_command(UDF_INIT *initid, UDF_ARGS *args, char *result,
                               unsigned long *length, char *is_null, char *error)
 {
   CommandInfo *info = (CommandInfo *)initid->ptr;
-  grn_ctx *ctx = &(info->ctx);
+  grn_ctx *ctx = info->ctx;
   char *command;
   unsigned int command_length;
   int flags = 0;
@@ -152,13 +164,12 @@ MRN_API void mroonga_command_deinit(UDF_INIT *initid)
 {
   CommandInfo *info = (CommandInfo *)initid->ptr;
   if (info) {
-    grn_obj *db = grn_ctx_db(&(info->ctx));
-    if (db) {
-      grn_obj_close(&(info->ctx), db);
+    if (!info->use_shared_db) {
+      grn_obj_close(info->ctx, info->db);
     }
-    grn_ctx_fin(&(info->ctx));
-    info->result.free();
-    my_free(info, MYF(0));
+    mrn_context_pool->release(info->ctx);
+    MRN_STRING_FREE(info->result);
+    my_free(info);
   }
 }
 
